Package: BayesPPR
Title: Bayesian Projection Pursuit Regression
Version: 0.1.0
Authors@R: c(
    person(given = "Gavin", 
           family = "Collins", 
           role = c("aut", "ctb"), 
           email = "gqcolli@sandia.gov"),
    person(given = "J. Derek",
           family = "Tucker",
           role = c("ctb", "cre"),
           email = "jdtuck@sandia.gov",
           comment = c(ORCID = "0000-0001-8844-2169")))
Description: Bayesian fitting of projection pursuit regression model. 
  Built to handle continuous and categorical inputs and
  scalar output (Collins et al., 2023 <DOI:10.1007/s11222-023-10334-z>).
License: MIT + file LICENSE
Encoding: UTF-8
Imports: methods
Suggests: testthat (>= 3.0.0)
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
URL: https://github.com/gqcollins/BayesPPR
BugReports: https://github.com/gqcollins/BayesPPR/issues
NeedsCompilation: no
Packaged: 2026-05-12 21:35:37 UTC; jdtuck
Author: Gavin Collins [aut, ctb],
  J. Derek Tucker [ctb, cre] (ORCID:
    <https://orcid.org/0000-0001-8844-2169>)
Maintainer: J. Derek Tucker <jdtuck@sandia.gov>
Repository: CRAN
Date/Publication: 2026-05-18 18:10:21 UTC
Built: R 4.6.0; ; 2026-05-18 20:10:18 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(plot,bppr)
S3method(predict,bppr)
S3method(predict,bppr_pca)
S3method(print,bppr)
S3method(summary,bppr)
export(bppr)
export(bppr_pca)
export(bppr_resume)
import(graphics)
import(stats)
import(utils)

# BayesPPR

<!-- badges: start -->
[![R-CMD-check](https://github.com/gqcollins/BayesPPR//actions/workflows/r.yml/badge.svg)](https://github.com/gqcollins/BayesPPR//actions/workflows/r.yml)
[![CRAN status](https://www.r-pkg.org/badges/version/BayesPPR)](https://CRAN.R-project.org/package=BayesPPR)
<!-- badges: end -->


## Description

BayesPPR is an R package used for fitting a Bayesian Projection Pursuit Regression model. (Collins et al., 2023 <DOI:10.1007/s11222-023-10334-z>)

### Installation
------------------------------------------------------------------------------
v0.1.0 is on [CRAN](https://cran.r-project.org/package=BayesPPR) and can
be installed as

``` r
install.packages("BayesPPR")`
```

For a more up to date, but may not be stable version from git
repository.

1.  Download zip or tar.gz of package or clone repository
2.  Install into R (\> 4.3.0)

``` r
library(devtools)
install_github("gqcollins/BayesPPR")
```

------------------------------------------------------------------------------


## Attribution

Copyright 2022 for C21079 BayesPPR 

This program is open source under the BSD-3 License. Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this list of conditions and
the following disclaimer.
 
2.Redistributions in binary form must reproduce the above copyright notice, this list of conditions
and the following disclaimer in the documentation and/or other materials provided with the
distribution.
 
3.Neither the name of the copyright holder nor the names of its contributors may be used to endorse
or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


© 2022. Triad National Security, LLC. All rights reserved. This program was produced under U.S. Government contract 89233218CNA000001 for Los Alamos National Laboratory (LANL), which is operated by Triad National Security, LLC for the U.S. Department of Energy/National Nuclear Security Administration. All rights in the program are reserved by Triad National Security, LLC, and the U.S. Department of Energy/National Nuclear Security Administration. The Government is granted for itself and others acting on its behalf a nonexclusive, paid-up, irrevocable worldwide license in this material to reproduce, prepare derivative works, distribute copies to the public, perform publicly and display publicly, and to permit others to do so.

\donttest{
  ################################
  ### univariate example
  ################################
  ## simulate data (Friedman function)
  f <- function(X){
    10*sin(pi*X[,1]*X[,2]) + 20*(X[,3] - .5)^2 + 10*X[,4] + 5*X[,5]
  }
  n <- 500 # number of observations
  p <- 10 #10 variables, only first 5 matter
  X <- matrix(runif(n*p), n, p)
  y <- f(X) + rnorm(n)

  ## fit BPPR
  fit <- bppr(X, y)

  ## plot diagnostics
  plot(fit)

  ## get out-of-sample predictions
  X_test <- matrix(runif(n*p), n, p)
  preds <- predict(fit, X_test) # posterior predictive samples
}

## minimal example for CRAN testing
fit <- bppr(matrix(1:50), 1:50, n_post = 2, n_burn = 0, n_adapt = 0)

pca_setup <- function(X, Y, n_pc = NULL, prop_var = 0.99){
  if(prop_var > 1  ||  prop_var < 0){
    stop('prop_var must be between 0 and 1')
  }

  n <- nrow(X)
  Y <- as.matrix(Y)

  if(nrow(Y) == n){
    D <- ncol(Y)
  }else{
    if(ncol(Y) == n){
      Y <- t(Y)
      D <- ncol(Y)
    }else{
      stop('X, Y dimension mismatch')
    }
  }

  if(D == 1){
    stop('univariate Y: use bppr instead of bppr_pca')
  }

  if(n == D){
    warning("Caution: because Y is square, please ensure that each row of X corresponds to a row of Y (and not a column)")
  }

  if(!is.null(n_pc)){
    if(n_pc > max(n, D)){
      n_pc <- max(n, D)
      warning('n_pc too large, using all PCs intead')
    }
  }

  mn_Y <- sd_Y <- numeric(D)
  Y_st <- matrix(nrow = n, ncol = D)
  for(k in 1:D){
    mn_Y[k] <- mean(Y[, k])
    sd_Y[k] <- sd(Y[, k])
    Y_st[, k] <- (Y[, k] - mn_Y[k]) / sd_Y[k]
  }

  svd_Y <- svd(t(Y_st))
  ev_Y <- svd_Y$d^2

  if(is.null(n_pc)){
    n_pc <- which(cumsum(ev_Y / sum(ev_Y)) > prop_var)[1]
  }

  basis_Y <- svd_Y$u[, 1:n_pc, drop=FALSE] %*% diag(svd_Y$d[1:n_pc], nrow = n_pc) # columns are basis functions
  Y_new <- svd_Y$v[, 1:n_pc, drop=FALSE]

  trunc_error <- Y_new %*% t(basis_Y) - Y_st

  out <- list(basis_Y = basis_Y, Y_new = Y_new, trunc_error = trunc_error, ev_Y = ev_Y,
              X = X, Y = Y, mn_Y = mn_Y, sd_Y = sd_Y, n_pc = n_pc, prop_var = prop_var)

  return(out)
}

pca_reverse <- function(preds_Y_new, pca_Y){
  pca_Y$basis_Y %*% t(preds_Y_new) * pca_Y$sd_Y + pca_Y$mn_Y
}

relu <- function(x){ # Get relu of input
  (abs(x) + x) / 2
}

get_mns_basis <- function(u, knots){ # Get modified natural spline basis
  n_knots <- length(knots)
  df <- n_knots - 2

  basis <- relu(u - knots[1]) # Initialize basis -- would be basis <- u for ns

  if(df > 1){
    n_internal_knots <- n_knots - 3
    r <- d <- list() # Temporary lists for computing basis
    for(k in 1:(n_knots - 1)){
      r[[k]] <- relu(u - knots[k + 1])^3
    }
    for(k in 1:df){
      d[[k]] <- (r[[k]] - r[[n_knots - 1]]) / (knots[n_knots] - knots[k + 1])
    }
    for(k in 1:n_internal_knots){
      basis <- cbind(basis, d[[k]] - d[[df]])
    }
  }

  basis
}

get_cat_basis <- function(X){
  if(ncol(X) > 1){
    tp <- (1 - X[, 1])
    for(j in 2:ncol(X)){
      tp <- tp * (1 - X[, j])
    }
    basis <- matrix(1 - tp)
  }else{
    basis <- X
  }
  basis
}

########################################################################
## prediction function
########################################################################

#' @title BayesPPR Predictions
#'
#' @description Predict function for BayesPPR. Outputs posterior predictive samples for the desired MCMC iterations.
#' @param object a fitted model, output from the \code{bppr} function.
#' @param newdata a matrix of new input values at which to predict.  The columns should correspond to the same variables used in the \code{bppr} function.
#' @param idx_use index of Markov samples to use when generating predictions.
#' @param ... further arguments passed to or from other methods.
#' @details Bare-bones methods. Could be improved for efficiency.
#' @return A matrix with the same number of rows as \code{newdata} and columns corresponding to all MCMC iterations indexed by idx_use. These are samples from the posterior predictive distribution.
#' @seealso \link{bppr} for model fitting.
#' @export
#' @examples
#' # See examples in bppr documentation.
#'
predict.bppr <- function(object, newdata, idx_use = NULL, ...){
  newdata <- as.matrix(newdata)
  n <- nrow(newdata)
  p <- ncol(newdata)
  mn_X <- object$mn_X
  sd_X <- object$sd_X
  for(j in 1:p){
    newdata[, j] <- (newdata[, j] - mn_X[j]) / sd_X[j]
  }
  coefs <- object$coefs
  proj_dir <- object$proj_dir
  n_ridge <- object$n_ridge
  n_act <- object$n_act
  feat <- object$feat
  knots <- object$knots
  df_spline <- object$df_spline
  n_keep <- object$n_keep

  if(is.null(idx_use)){
    idx_use <- 1:n_keep
  }else if(max(idx_use) > n_keep){
    stop("invalid 'idx_use'")
  }
  n_use <- length(idx_use)

  ridge_basis <- vector('list', n_ridge[idx_use[1]])

  preds <- matrix(0, nrow = n_use, ncol = n)
  for(i in 1:n_use){
    preds[i, ] <- coefs[[idx_use[i]]][1]
    calc_all_bases <- (i == 1)
    if(n_ridge[idx_use[i]] > 0){
      if(!calc_all_bases  &&  n_ridge[idx_use[i]] != n_ridge[idx_use[i-1]]){
        calc_all_bases <- TRUE
        ridge_basis <- vector('list', n_ridge[idx_use[i]])
      }
      basis_idx_start <- 2
      for(j in 1:n_ridge[idx_use[i]]){
        if(is.na(knots[[idx_use[i]]][[j]][1])){ # No continuous features in this basis
          basis_idx <- basis_idx_start
          basis_idx_start <- basis_idx_start + 1
          if(is.na(proj_dir[[idx_use[i]]][[j]][1])){ # all categorical features in this basis
            if(calc_all_bases){
              ridge_basis[[j]] <- get_cat_basis(newdata[, feat[[idx_use[i]]][[j]], drop = FALSE])
            }
          }else{ # some discrete quantitative features in this basis
            if(calc_all_bases  ||
               n_act[[idx_use[i]]][j] != n_act[[idx_use[i-1]]][j]  ||
               any(proj_dir[[idx_use[i]]][[j]] != proj_dir[[idx_use[i-1]]][[j]])){
              ridge_basis[[j]] <- newdata[, feat[[idx_use[i]]][[j]], drop = FALSE] %*% proj_dir[[idx_use[i]]][[j]]
            }
          }
        }else{ # At least one continuous feature in this basis
          basis_idx <- basis_idx_start:(basis_idx_start + df_spline - 1)
          basis_idx_start <- basis_idx_start + df_spline
          if(calc_all_bases  ||
             n_act[[idx_use[i]]][j] != n_act[[idx_use[i-1]]][j]  ||
             knots[[idx_use[i]]][[j]][1] != knots[[idx_use[i-1]]][[j]][1]){
            proj <- newdata[, feat[[idx_use[i]]][[j]], drop = FALSE] %*% proj_dir[[idx_use[i]]][[j]]
            ridge_basis[[j]] <- get_mns_basis(proj, knots[[idx_use[i]]][[j]]) # Get basis function
          }
        }
        preds[i, ] <- preds[i, ] + ridge_basis[[j]] %*% coefs[[idx_use[i]]][basis_idx] # Get predictions for jth basis function
      }
    }
  }

  return(preds)
}

########################################################################
## Resume the MCMC given a current bppr object
########################################################################

#' @title Resume Sampling of BPPR Parameters
#'
#' @description Resumes fitting a BayesPPR model where the last Markov chain left off, using the same model parameters.
#' @param object an object of class \code{"bppr"} containing the values of the current Markov chain, the latest of which constitute initial values for the new chain.
#' @param append logical; if \code{TRUE}, new samples will be appended to samples in \code{object}. Otherwise, output will only include new samples.
#' @param n_post number of posterior draws to obtain from the Markov chain after burn-in.
#' @param n_burn number of draws to burn before obtaining \code{n_post} draws for inference. If \code{prior_coefs == "flat"} then these are absorbed into the adapt phase.
#' @param n_adapt number of adaptive MCMC iterations to perform before burn-in. Skips sampling basis coefficients and residual variance to save time.
#' @param n_thin keep every n_thin posterior draws after burn-in.
#' @param print_every print the iteration number every print_every iterations. Use \code{print_every = 0} to silence.
#' @details Continues exploring BayesPPR model space using RJMCMC, beginning at the values contained in the last iteration of \code{object}.
#' @return An object of class \code{"bppr"}. Predictions can be obtained by passing the entire object to the \code{predict.bppr} function.
#' @keywords nonparametric projection pursuit regression splines
#' @seealso \link{predict.bppr} for prediction.
#' @export
#' @import stats
#' @import utils
bppr_resume <- function(object, append = FALSE, n_post = 1000, n_burn = 9000, n_adapt = 0, n_thin = 1, print_every = 1000){

  if(methods::is(object,"bppr")){
    n_keep <- object$n_keep
    if(object$prior_coefs == 'zs'){
      var_coefs_init <- object$var_coefs[n_keep]
    }else{
      var_coefs_init <- NULL
    }

    bppr_init <- list(n_ridge = object$n_ridge[[n_keep]], n_act = object$n_act[[n_keep]],
                      feat = object$feat[[n_keep]], proj_dir = object$proj_dir[[n_keep]],
                      knots = object$knots[[n_keep]], coefs = object$coefs[[n_keep]],
                      sd_resid = object$sd_resid[n_keep], var_coefs = var_coefs_init,
                      w_n_act = object$w_n_act, w_feat = object$w_feat)

    fit <- bppr(X = object$X, y = object$y, n_ridge_mean = object$n_ridge_mean,
                n_ridge_max = object$n_ridge_max, n_act_max = object$n_act_max,
                df_spline = object$df_spline, prob_relu = object$prob_relu,
                prior_coefs = object$prior_coefs, shape_var_coefs = object$shape_var_coefs,
                scale_var_coefs = object$scale_var_coefs, n_dat_min = object$n_dat_min,
                scale_proj_dir_prop = object$scale_proj_dir_prop,
                adapt_act_feat = object$adapt_act_feat, n_post = n_post, n_burn = n_burn,
                n_adapt = n_adapt, n_thin = n_thin, print_every = print_every,
                bppr_init = bppr_init)

    if(append){
      fit$n_ridge <- c(object$n_ridge, fit$n_ridge)
      fit$n_act <- c(object$n_act, fit$n_act)
      fit$feat <- c(object$feat, fit$feat)
      fit$proj_dir <- c(object$proj_dir, fit$proj_dir)
      fit$knots <- c(object$knots, fit$knots)
      fit$coefs <- c(object$coefs, fit$coefs)
      fit$sd_resid <- c(object$sd_resid, fit$sd_resid)
      fit$var_coefs <- c(object$var_coefs, fit$var_coefs)
      fit$n_keep <- object$n_keep + fit$n_keep
      fit$n_post <- c(object$n_post, fit$n_post)
      fit$n_burn <- c(object$n_burn, fit$n_burn)
      fit$n_adapt <- c(object$n_adapt, fit$n_adapt)
    }
  }else if(methods::is(object,"bppr_pca")){
    n_keep <- object$fit_list[[1]]$n_keep

    bppr_init_list <- lapply(1:object$pca_Y$n_pc, function(k){
      if(object$fit_list[[k]]$prior_coefs == 'zs'){
        var_coefs_init <- object$fit_list[[k]]$var_coefs[n_keep]
      }else{
        var_coefs_init <- NULL
      }

      list(n_ridge = object$fit_list[[k]]$n_ridge[[n_keep]],
           n_act = object$fit_list[[k]]$n_act[[n_keep]], feat = object$fit_list[[k]]$feat[[n_keep]],
           proj_dir = object$fit_list[[k]]$proj_dir[[n_keep]], knots = object$fit_list[[k]]$knots[[n_keep]],
           coefs = object$fit_list[[k]]$coefs[[n_keep]], sd_resid = object$fit_list[[k]]$sd_resid[n_keep],
           var_coefs = var_coefs_init, w_n_act = object$fit_list[[k]]$w_n_act,
           w_feat = object$fit_list[[k]]$w_feat)
    })

    fit <- bppr_pca(X = object$pca_Y$X, Y = object$pca_Y$Y, n_pc = object$pca_Y$n_pc,
                    prop_var = object$pca_Y$prop_var, n_cores = object$n_cores, par_type = object$par_type,
                    n_ridge_mean = object$fit_list[[1]]$n_ridge_mean, n_ridge_max = object$fit_list[[1]]$n_ridge_max,
                    n_act_max = object$fit_list[[1]]$n_act_max, df_spline = object$fit_list[[1]]$df_spline,
                    prob_relu = object$fit_list[[1]]$prob_relu, prior_coefs = object$fit_list[[1]]$prior_coefs,
                    shape_var_coefs = object$fit_list[[1]]$shape_var_coefs, scale_var_coefs = object$fit_list[[1]]$scale_var_coefs,
                    n_dat_min = object$fit_list[[1]]$n_dat_min, scale_proj_dir_prop = object$fit_list[[1]]$scale_proj_dir_prop,
                    adapt_act_feat = object$fit_list[[1]]$adapt_act_feat, n_post = n_post, n_burn = n_burn, n_adapt = n_adapt,
                    n_thin = n_thin, print_every = print_every, bppr_init_list = bppr_init_list)

    if(append){
      for(k in 1:object$pca_Y$n_pc){
        fit$fit_list[[k]]$n_ridge <- c(object$fit_list[[k]]$n_ridge, fit$fit_list[[k]]$n_ridge)
        fit$fit_list[[k]]$n_act <- c(object$fit_list[[k]]$n_act, fit$fit_list[[k]]$n_act)
        fit$fit_list[[k]]$feat <- c(object$fit_list[[k]]$feat, fit$fit_list[[k]]$feat)
        fit$fit_list[[k]]$proj_dir <- c(object$fit_list[[k]]$proj_dir, fit$fit_list[[k]]$proj_dir)
        fit$fit_list[[k]]$knots <- c(object$fit_list[[k]]$knots, fit$fit_list[[k]]$knots)
        fit$fit_list[[k]]$coefs <- c(object$fit_list[[k]]$coefs, fit$fit_list[[k]]$coefs)
        fit$fit_list[[k]]$sd_resid <- c(object$fit_list[[k]]$sd_resid, fit$fit_list[[k]]$sd_resid)
        fit$fit_list[[k]]$var_coefs <- c(object$fit_list[[k]]$var_coefs, fit$fit_list[[k]]$var_coefs)
        fit$fit_list[[k]]$n_keep <- object$fit_list[[k]]$n_keep + fit$fit_list[[k]]$n_keep
        fit$fit_list[[k]]$n_post <- c(object$fit_list[[k]]$n_post, fit$fit_list[[k]]$n_post)
        fit$fit_list[[k]]$n_burn <- c(object$fit_list[[k]]$n_burn, fit$fit_list[[k]]$n_burn)
        fit$fit_list[[k]]$n_adapt <- c(object$fit_list[[k]]$n_adapt, fit$fit_list[[k]]$n_adapt)
      }
    }
  }else{
    stop("object must be of class 'bppr' or 'bppr_pca'")
  }

  return(fit)
}

dwallenius <- function(feat_w, feat){ # Wallenius non-central t
  feat_w <- feat_w / sum(feat_w)
  alpha <- feat_w[feat] / sum(feat_w[-feat])
  j <- length(alpha)
  ss <- 1 + (-1)^j * 1/(sum(alpha) + 1)
  if(j > 1){
    for(i in 1:(j - 1))
      ss <- ss + (-1)^(i) * sum(1/(colSums(combn(alpha, i)) + 1))
  }
  ss
}

########################################################################
## BayesPPR for multivariate response
########################################################################

#' @title BayesPPR for Multivariate Response
#'
#' @description Fits a BayesPPR model to multivariate response using RJMCMC. Can handle categorical features.
#' @param X a data frame or matrix of predictors. Categorical features should be coded as numeric.
#' @param Y a data frame or matrix of responses, with columns representing different dimensions of the response.
#' @param n_pc number of principle components to be used in pca approximation of Y.
#' @param prop_var proportion of variance to be explained by the first \code{n_pc} principle components. Used for automatic selection of \code{n_pc} if \code{is.null(n_pc)}.
#' @param n_cores number of cores the user desired to utilize for parallel computation.
#' @param par_type a character variable dicatating the type of parallel computation to be performed. Supported values include \code{"fork"}, which uses \code{parallel::mclapply()}, and \code{"socket"}, which uses \code{parallel::parLapply()}.
#' @param n_ridge_mean mean for Poisson prior on the number of ridge functions.
#' @param n_ridge_max maximum number of ridge functions allowed in the model. Used to avoid memory overload. Defaults to 150 unless the number of observed responses is small.
#' @param n_act_max maximum number of active variables in any given ridge function. Defaults to 3 unless categorical features are detected, in which case the default is larger.
#' @param df_spline degrees of freedom for spline basis. Stability should be examined for anything other than 4.
#' @param prob_relu prior probability that any given ridge function uses a relu transformation.
#' @param prior_coefs form of the prior distribution for the basis coefficients. Default is \code{"zs"} for the Zellner-Siow prior. The other option is \code{"flat"}, which is an improper prior.
#' @param shape_var_coefs shape for IG prior on the variance of the basis function coefficients. Default is for the Zellner-Siow prior. For the flat, improper prior, \code{shape_var_coefs} is ignored.
#' @param scale_var_coefs scale for IG prior on the variance of the basis function coefficients. Default is for the Zellner-Siow prior. For the flat, improper prior, \code{scale_var_coefs} is ignored.
#' @param n_dat_min minimum number of observed non-zero datapoints in a ridge function. Defaults to 20 or 0.1 times the number of observations, whichever is smaller.
#' @param scale_proj_dir_prop scale parameter for generating proposed projection directions. Should be in (0, 1); default is about 0.002.
#' @param adapt_act_feat logical; if \code{TRUE}, use adaptive proposal for feature index sets and number of active features.
#' @param w_n_act vector of weights for number of active variables in a ridge function, used in generating proposed basis functions. If \code{adapt_act_feat == FALSE}, it is also used for the prior distribution. Default is \code{rep(1, n_act_max)}.
#' @param w_feat vector of weights for feature indices used in a ridge function, used in generating proposed basis functions. If \code{adapt_act_feat == FALSE}, it is also used for the prior distribution. Default is \code{rep(1, ncol(X))}.
#' @param n_post number of posterior draws to obtain from the Markov chain after burn-in.
#' @param n_burn number of draws to burn before obtaining \code{n_post} draws for inference. If \code{prior_coefs == "flat"} then these are absorbed into the adapt phase.
#' @param n_adapt number of adaptive MCMC iterations to perform before burn-in. Skips sampling basis coefficients and residual variance to save time.
#' @param n_thin keep every n_thin posterior draws after burn-in.
#' @param print_every print the iteration number every print_every iterations. Use \code{print_every = 0} to silence.
#' @param bppr_init_list list of length \code{n_pc}, each element containing a list initial values for the Markov chain. Used by \link{bppr_resume}.
#' @details Explores BayesPPR model space using RJMCMC. The BayesPPR model has \deqn{y = f(x) + \epsilon,  ~~\epsilon \sim N(0,\sigma^2)} \deqn{f(x) = \beta_0 + \sum_{j=1}^M \beta_j B_j(x)} and \eqn{B_j(x)} is a natural spline basis expansion. We use priors \deqn{\beta \sim N(0,\sigma^2/\tau (B'B)^{-1})} \deqn{M \sim Poisson(\lambda)} as well as the hyper-prior on the variance \eqn{\tau} of the coefficients \eqn{\beta} mentioned in the arguments above.
#' @return An object of class \code{"bppr_pca"}. Predictions can be obtained by passing the entire object to the \code{predict.bppr_pca} function.
#' @keywords nonparametric projection pursuit regression splines principle component analysis
#' @seealso \link{predict.bppr_pca} for prediction.
#' @export
#' @import stats
#' @import utils
#' @examples
#' # See examples in bppr documentation.
#'
bppr_pca <- function(X, Y, n_pc = NULL, prop_var = 0.99, n_cores = 1, par_type = 'fork', n_ridge_mean = 10, n_ridge_max = NULL, n_act_max = NULL, df_spline = 4, prob_relu = 2/3, prior_coefs = "zs", shape_var_coefs = NULL, scale_var_coefs = NULL, n_dat_min = NULL, scale_proj_dir_prop = NULL, adapt_act_feat = TRUE, w_n_act = NULL, w_feat = NULL, n_post = 1000, n_burn = 9000, n_adapt = 0, n_thin = 1, print_every = NULL, bppr_init_list = NULL){

  pca_Y <- pca_setup(X, Y, n_pc = n_pc, prop_var = prop_var)
  n_pc <- pca_Y$n_pc

  n_cores_max <- parallel::detectCores()

  if(n_cores > n_cores_max){
    warning(paste0("Specified n_cores = ", n_cores, ". Proceeding with n_cores = min(n_cores, n_pc, detectCores()) = ",
                   min(n_cores, n_pc, n_cores_max)))
  }
  n_cores <- min(n_cores, n_pc, n_cores_max)

  if(is.null(print_every)){
    if(n_cores == 1){
      print_every <- 1000
    }else{
      print_every <- 0
    }
  }

  if(is.null(bppr_init_list)){
    bppr_init_code <- "NULL"
  }else{
    bppr_init_code <- "bppr_init_list[[i]]"
  }

  run_bppr <- parse(text = paste0(
  "bppr(X, pca_Y$Y_new[, i], n_ridge_mean = n_ridge_mean, n_ridge_max = n_ridge_max,
        n_act_max = n_act_max, df_spline = df_spline, prob_relu = prob_relu,
        prior_coefs = prior_coefs, shape_var_coefs = shape_var_coefs,
        scale_var_coefs = scale_var_coefs, n_dat_min = n_dat_min,
        scale_proj_dir_prop = scale_proj_dir_prop, adapt_act_feat = adapt_act_feat,
        w_n_act = w_n_act, w_feat = w_feat, n_post = n_post, n_burn = n_burn,
        n_adapt = n_adapt, n_thin = n_thin, print_every = print_every,
        bppr_init = ", bppr_init_code, ")")
  )

  if(n_cores == 1){
    fit_list <- lapply(1:n_pc, function(i) eval(run_bppr))
  }else if(par_type == "socket"){
    cl <- parallel::makeCluster(n_cores, setup_strategy = "sequential")
    fit_list <- parallel::parLapply(cl, 1:n_pc, function(i) eval(run_bppr))
    parallel::stopCluster(cl)
  }else if(par_type == "fork"){
    fit_list <- parallel::mclapply(1:pca_Y$n_pc, function(i) eval(run_bppr),
      mc.cores = n_cores, mc.preschedule = FALSE)
  }

 structure(list(pca_Y = pca_Y, fit_list = fit_list,
                n_cores = n_cores, par_type = par_type),
           class = 'bppr_pca')
}

########################################################################
## main BayesPPR function
########################################################################

#' @title Bayesian Projection Pursuit Regression
#'
#' @description Fits a BayesPPR model using RJMCMC. Can handle categorical features.
#' @param X a data frame or matrix of predictors. Categorical features should be coded as numeric.
#' @param y a numeric response vector.
#' @param n_ridge_mean mean for Poisson prior on the number of ridge functions.
#' @param n_ridge_max maximum number of ridge functions allowed in the model. Used to avoid memory overload. Defaults to 150 unless the number of observed responses is small.
#' @param n_act_max maximum number of active variables in any given ridge function. Defaults to 3 unless categorical features are detected, in which case the default is larger.
#' @param df_spline degrees of freedom for spline basis. Stability should be examined for anything other than 4.
#' @param prob_relu prior probability that any given ridge function uses a relu transformation.
#' @param prior_coefs form of the prior distribution for the basis coefficients. Default is \code{"zs"} for the Zellner-Siow prior. The other option is \code{"flat"}, which is an improper prior.
#' @param shape_var_coefs shape for IG prior on the variance of the basis function coefficients. Default is for the Zellner-Siow prior. For the flat, improper prior, \code{shape_var_coefs} is ignored.
#' @param scale_var_coefs scale for IG prior on the variance of the basis function coefficients. Default is for the Zellner-Siow prior. For the flat, improper prior, \code{scale_var_coefs} is ignored.
#' @param n_dat_min minimum number of observed non-zero datapoints in a ridge function. Defaults to 20 or 0.1 times the number of observations, whichever is smaller.
#' @param scale_proj_dir_prop scale parameter for generating proposed projection directions. Should be in (0, 1); default is about 0.002.
#' @param adapt_act_feat logical; if \code{TRUE}, use adaptive proposal for feature index sets and number of active features.
#' @param w_n_act vector of weights for number of active variables in a ridge function, used in generating proposed basis functions. If \code{adapt_act_feat == FALSE}, it is also used for the prior distribution. Default is \code{rep(1, n_act_max)}.
#' @param w_feat vector of weights for feature indices used in a ridge function, used in generating proposed basis functions. If \code{adapt_act_feat == FALSE}, it is also used for the prior distribution. Default is \code{rep(1, ncol(X))}.
#' @param n_post number of posterior draws to obtain from the Markov chain after burn-in.
#' @param n_burn number of draws to burn before obtaining \code{n_post} draws for inference. If \code{prior_coefs == "flat"} then these are absorbed into the adapt phase.
#' @param n_adapt number of adaptive MCMC iterations to perform before burn-in. Skips sampling basis coefficients and residual variance to save time.
#' @param n_thin keep every n_thin posterior draws after burn-in.
#' @param print_every print the iteration number every print_every iterations. Use \code{print_every = 0} to silence.
#' @param bppr_init list of initial values for the Markov chain. Used by \link{bppr_resume}.
#' @details Explores BayesPPR model space using RJMCMC. The BayesPPR model has \deqn{y = f(x) + \epsilon,  ~~\epsilon \sim N(0,\sigma^2)} \deqn{f(x) = \beta_0 + \sum_{j=1}^M \beta_j B_j(x)} and \eqn{B_j(x)} is a natural spline basis expansion. We use priors \deqn{\beta \sim N(0,\sigma^2/\tau (B'B)^{-1})} \deqn{M \sim Poisson(\lambda)} as well as the hyper-prior on the variance \eqn{\tau} of the coefficients \eqn{\beta} mentioned in the arguments above.
#' @return An object of class \code{"bppr"}. Predictions can be obtained by passing the entire object to the \code{predict.bppr} function.
#' @keywords nonparametric projection pursuit regression splines
#' @seealso \link{predict.bppr} for prediction.
#' @export
#' @import stats
#' @import utils
#' @example inst/examples.R
#'
bppr <- function(X, y, n_ridge_mean = 10, n_ridge_max = NULL, n_act_max = NULL, df_spline = 4, prob_relu = 2/3, prior_coefs = "zs", shape_var_coefs = NULL, scale_var_coefs = NULL, n_dat_min = NULL, scale_proj_dir_prop = NULL, adapt_act_feat = TRUE, w_n_act = NULL, w_feat = NULL, n_post = 1000, n_burn = 9000, n_adapt = 0, n_thin = 1, print_every = 1000, bppr_init = NULL){
  # Manage posterior draws
  if(n_thin > n_post){
    stop('n_thin > n_post. No posterior samples will be obtained.')
  }
  n_post <- n_post - n_post %% n_thin
  n_keep <- n_post/n_thin
  n_pre <- n_adapt + n_burn
  n_draws <- n_pre + n_post
  idx <- c(rep(1, n_pre), rep(1:n_keep, each = n_thin))
  if(prior_coefs == 'flat'){
    n_adapt <- n_pre
    n_burn <- 0
  }

  # Pre-processing
  n <- length(y)
  p <- ncol(X)

  if(is.null(w_feat)){
    w_feat <- rep(1, p)
  }

  mn_X <- sd_X <- numeric(p)
  X_st <- X
  feat_type <- character(p)
  for(j in 1:p){
    n_unique <- length(unique(X[, j]))
    if(n_unique <= 2){
      mn_X[j] <- 0
      sd_X[j] <- 1
      if(n_unique == 1){
        feat_type[j] <- ''
        w_feat[j] <- 0
      }else{
        feat_type[j] <- 'cat'
      }
    }else{
      mn_X[j] <- mean(X[, j])
      sd_X[j] <- sd(X[, j])
      X_st[, j] <- (X[, j] - mn_X[j]) / sd_X[j]
      if(n_unique <= df_spline){
        feat_type[j] <- 'disc'
      }else{
        feat_type[j] <- 'cont'
      }
    }
  }

  if(is.null(n_act_max)){
    n_cat <- sum(feat_type == 'cat')
    n_act_max <- min(3, p - n_cat) + min(3, ceiling(n_cat/2))
  }

  if(is.null(w_n_act)){
    w_n_act <- rep(1, n_act_max)
  }

  if(is.null(scale_proj_dir_prop)){
    proj_dir_prop_prec <- 1000 # scale_proj_dir_prop = 0.002
  }else if(scale_proj_dir_prop > 1  ||  scale_proj_dir_prop <= 0){
    stop("scale_proj_dir_prop must be in (0, 1]")
  }else{
    proj_dir_prop_prec <- 1/scale_proj_dir_prop
    proj_dir_prop_prec <- (proj_dir_prop_prec - 1) + sqrt(proj_dir_prop_prec * (proj_dir_prop_prec - 1))
  }

  if(is.null(n_dat_min)){
    n_dat_min <- min(20, 0.1 * n)
  }
  if(n_dat_min <= df_spline){
    warning('n_dat_min too small. If n_dat_min was set by default, df_spline is large compared to the sample size. Setting n_dat_min = df_spline + 1')
    n_dat_min <- df_spline + 1
  }
  p_dat_max <- 1 - n_dat_min / n # Maximum proportion of inactive datapoints in each ridge function

  if(is.null(n_ridge_max)){
    n_ridge_max <- min(150, floor(n/df_spline) - 2)
  }
  if(n_ridge_max <= 0){
    stop('n_ridge_max <= 0. If n_ridge_max was set by default, df_spline is too large compared to the sample size.')
  }

  knot_quants <- seq(0, 1, length.out = df_spline + 1) # Quantiles for knot locations (except initial knot)

  if(prior_coefs == 'zs'){
    if(is.null(shape_var_coefs)){
      shape_var_coefs <- 0.5
    }
    if(is.null(scale_var_coefs)){
      scale_var_coefs <- n/2
    }
    var_coefs <- numeric(n_keep)
  }else if(prior_coefs == 'flat'){
    var_coefs <- NA
    c_var_coefs <- 1
  }else{
    stop("prior_coefs must be either 'zs' or 'flat'")
  }
  sd_resid <- numeric(n_keep) # Error standard deviation
  coefs <- vector('list', n_keep) # Basis coefficients
  n_ridge <- numeric(n_keep) # Number of ridge functions
  n_act <- vector('list', n_keep) # Number of active features for jth ridge function
  feat <- vector('list', n_keep) # Features being used in jth ridge function
  knots <- vector('list', n_keep) # Location of knots for nsplines
  proj_dir <- vector('list', n_keep) # Ridge directions
  proj_dir_mn <- lapply(1:n_act_max, function(a) rep(1/sqrt(a), a)) # prior mean for proj_dir (arbitrary, since prior precision is zero)

  n_basis_ridge <- 1 # Number of basis functions in each ridge function (first ridge function is the intercept)
  ridge_type <- character()
  j_quant <- integer() # which ridge functions are quantitative?
  n_quant <- 0 # how many ridge functions are quantitative?
  basis_mat <- matrix(rep(1, n)) # Current basis matrix
  basis_idx <- list(1) # Current indices of segments of basis functions
  BtB <- matrix(nrow = n_ridge_max * df_spline + 1, ncol = n_ridge_max * df_spline + 1)
  Bty <- numeric(n_ridge_max * df_spline + 1)

  # Initialization
  if(is.null(bppr_init)){
    if(prior_coefs == 'zs'){
      var_coefs[1] <- scale_var_coefs / shape_var_coefs
      c_var_coefs <- var_coefs[1] / (var_coefs[1] + 1)
    }
    sd_resid[1] <- 1
    coefs[[1]] <- mean(y)
    n_act[[1]] <- numeric() # Number of active features for jth ridge function
    feat[[1]] <- list() # Features being used in jth ridge function
    knots[[1]] <- list() # Location of knots for nsplines
    proj_dir[[1]] <- list() # Ridge directions
  }else{
    if(prior_coefs == 'zs'){
      var_coefs[1] <- bppr_init$var_coefs
      c_var_coefs <- var_coefs[1] / (var_coefs[1] + 1)
    }
    sd_resid[1] <- bppr_init$sd_resid
    coefs[[1]] <- bppr_init$coefs
    n_ridge[1] <- bppr_init$n_ridge
    n_act[[1]] <- bppr_init$n_act
    feat[[1]] <- bppr_init$feat
    knots[[1]] <- bppr_init$knots
    proj_dir[[1]] <- bppr_init$proj_dir
    if(n_ridge[1] > 0){
      basis_idx_start <- 2
      for(j in 1:n_ridge[1]){
        if(any(feat_type[feat[[1]][[j]]] == 'cont')){
          ridge_type[j] <- 'cont'
          n_basis_ridge[j + 1] <- df_spline
          j_quant <- c(j_quant, j)
          proj <- X_st[, feat[[1]][[j]], drop = FALSE] %*% proj_dir[[1]][[j]]
          basis_mat <- cbind(basis_mat, get_mns_basis(proj, knots[[1]][[j]])) # Get basis function
          basis_idx[[j + 1]] <- basis_idx_start:(basis_idx_start + df_spline - 1)
          basis_idx_start <- basis_idx_start + df_spline
        }else{
          n_basis_ridge[j + 1] <- 1
          if(any(feat_type[feat[[1]][[j]]] == 'disc')){
            ridge_type[j] <- 'disc'
            n_basis_ridge[j + 1] <- 1
            j_quant <- c(j_quant, j)
            basis_mat <- cbind(basis_mat, X_st[, feat[[1]][[j]], drop = FALSE] %*% proj_dir[[1]][[j]])
            basis_idx[[j + 1]] <- basis_idx_start
            basis_idx_start <- basis_idx_start + 1
          }else{
            ridge_type[j] <- 'cat'
            n_basis_ridge[j] <- 1
            basis_mat <- cbind(basis_mat, get_cat_basis(X_st[, feat[[1]][[j]], drop = FALSE]))
            basis_idx[[j + 1]] <- basis_idx_start
            basis_idx_start <- basis_idx_start + 1
          }
        }
      }
      n_quant <- length(j_quant)
    }
    w_n_act <- bppr_init$w_n_act
    w_feat <- bppr_init$w_feat
  }

  n_basis_total <- sum(n_basis_ridge)
  BtB[1:n_basis_total, 1:n_basis_total] <- t(basis_mat) %*% basis_mat # inner product of basis_mat with itself
  Bty[1:n_basis_total] <- t(basis_mat) %*% y # inner product of basis_mat with y
  qf_info <- get_qf_info(BtB[1:n_basis_total, 1:n_basis_total, drop = FALSE], Bty[1:n_basis_total])
  if(n_adapt == 0) qf_info <- append_qf_inv_chol(qf_info, dim = n_basis_total)

  ssy <- c(t(y) %*% y) # Keep track of overall sse
  sse <- ssy - c_var_coefs * qf_info$qf
  log_mh_bd <- 0
  log_mh_act_feat <- 0

  if(n_adapt > 0){
    phase <- 'adapt'
  }else if(n_burn > 0){
    phase <- 'burn'
  }else{
    phase <- 'post-burn'
  }
  if(print_every > 0){
    start_time <- Sys.time()
    cat(paste0('MCMC iteration 1/', n_draws, ' (', phase, ') ',
               myTimestamp(), ' n ridge: ', n_ridge[1], '\n'))
    silent <- FALSE
  }else{
    print_every <- n_draws + 2
    silent <- TRUE
  }

  # Run MCMC
  if(n_draws > 1){
    for(it in 2:n_draws){
      # Set current it values to last it values (these will change during the iteration)
      if(idx[it] > idx[it - 1]){
        if(prior_coefs == 'zs') var_coefs[idx[it]] <- var_coefs[idx[it - 1]]
        sd_resid[idx[it]] <- sd_resid[idx[it - 1]]
        coefs[[idx[it]]] <- coefs[[idx[it - 1]]]
        n_ridge[idx[it]] <- n_ridge[idx[it - 1]]
        n_act[[idx[it]]] <- n_act[[idx[it - 1]]]
        feat[[idx[it]]] <- feat[[idx[it - 1]]]
        knots[[idx[it]]] <- knots[[idx[it - 1]]]
        proj_dir[[idx[it]]] <- proj_dir[[idx[it - 1]]]
      }

      if(it == n_adapt + 1){
        if(n_burn > 0){
          phase <- 'burn'
        }else{
          phase <- 'post-burn'
        }
      }
      if(it == n_pre + 1) phase <- 'post-burn'
      if((it - 1) %% print_every == 0  ||  ((it == n_adapt + 1  ||  it == n_pre + 1)  &&  !silent)){
        pr <- paste0('MCMC iteration ', it, '/', n_draws, ' (', phase, ') ',
                     myTimestamp(start_time), ' n ridge: ', n_ridge[idx[it]])
        cat(pr, '\n')
      }

      move_type <- get_move_type(n_ridge[idx[it]], n_quant, n_ridge_max)

      if(move_type == 'birth'){ # Birth step
        n_ridge_prop <- n_ridge[idx[it]] + 1
        n_act_prop <- sample(n_act_max, 1, prob = w_n_act) # Propose number of active features
        if(adapt_act_feat){
          log_mh_act_feat <- -(log(n_act_max) + log(w_n_act[n_act_prop]/sum(w_n_act))) # Nott, Kuk, Duc for n_act
          if(n_act_prop == 1){
            feat_prop <- sample(p, 1)
          }else{
            feat_prop <- sample(p, n_act_prop, prob = w_feat) # Propose features to include
            log_mh_act_feat <- log_mh_act_feat - (lchoose(p, n_act_prop) + log(dwallenius(w_feat, feat_prop))) # Nott, Kuk, Duc for feat
          }
        }else{
          feat_prop <- sample(p, n_act_prop, prob = w_feat) # Propose features to include
        }

        if(all(feat_type[feat_prop] == 'cat')){ # Are all of the proposed features categorical?
          ridge_type_prop <- 'cat'
          n_quant_prop <- n_quant
          proj_dir_prop <- knots_prop <- NA
          ridge_basis_prop <- get_cat_basis(X_st[, feat_prop, drop = FALSE])
          n_basis_prop <- 1
        }else{
          n_quant_prop <- n_quant + 1
          if(n_act_prop == 1){
            proj_dir_prop <- matrix(sample(c(-1, 1), 1))
          }else{
            proj_dir_prop <- rps(proj_dir_mn[[n_act_prop]], 0) # Propose direction
          }
          proj_prop <- X_st[, feat_prop, drop = FALSE] %*% proj_dir_prop # Get proposed projection

          if(any(feat_type[feat_prop] == 'cont')){ # Are any proposed features continuous?
            ridge_type_prop <- 'cont'
            max_knot0 <- quantile(proj_prop, p_dat_max)
            rg_knot0 <- (max_knot0 - min(proj_prop)) / prob_relu
            knot0_prop <- max_knot0 - rg_knot0 * runif(1)
            knots_prop <- c(knot0_prop, quantile(proj_prop[proj_prop > knot0_prop], knot_quants)) # Get proposed knots
            ridge_basis_prop <- get_mns_basis(proj_prop, knots_prop) # Get proposed basis functions
            n_basis_prop <- df_spline
          }else{ # The proposed features are a mix of categorical and discrete quantitative
            ridge_type_prop <- 'disc'
            knots_prop <- NA
            ridge_basis_prop <- proj_prop
            n_basis_prop <- 1
          }
        }

        PtP <- t(ridge_basis_prop) %*% ridge_basis_prop # inner product of proposed new basis functions
        BtP <- t(basis_mat) %*% ridge_basis_prop
        Pty <- t(ridge_basis_prop) %*% y

        j_birth <- n_ridge[idx[it]] + 1
        basis_idx_start <- basis_idx[[j_birth]][n_basis_ridge[j_birth]]
        basis_idx_prop <- (basis_idx_start + 1):(basis_idx_start + n_basis_prop)
        BtB[1:n_basis_total, basis_idx_prop] <- BtP
        BtB[basis_idx_prop, 1:n_basis_total] <- t(BtP)
        BtB[basis_idx_prop, basis_idx_prop] <- PtP
        Bty[basis_idx_prop] <- Pty

        qf_info_prop <- get_qf_info(BtB[1:(n_basis_total + n_basis_prop), 1:(n_basis_total + n_basis_prop)], Bty[1:(n_basis_total + n_basis_prop)])
        log_mh_bd_prop <- get_log_mh_bd(n_ridge_prop, n_quant_prop, n_ridge_max)

        if(!is.null(qf_info_prop)){
          if(qf_info_prop$qf < ssy){
            sse_prop <- ssy - c_var_coefs * qf_info_prop$qf

            # Compute the acceptance probability
            log_mh <- log_mh_bd - log_mh_bd_prop + log_mh_act_feat + # Adjustment for probability of birth proposal
              -n/2 * (log(sse_prop) - log(sse)) + # Part of the marginal likelihood
              log(n_ridge_mean/(n_ridge[idx[it]] + 1)) # Prior and proposal distribution
            if(prior_coefs == 'zs'){
              log_mh <- log_mh - n_basis_prop/2 * log(var_coefs[idx[it]] + 1) # The rest of the marginal likelihood for Zellner-Siow prior
            }else{
              log_mh <- log_mh + log(10e-6) # The rest of the marginal likelihood for flat prior
            }

            if(log(runif(1)) < log_mh){ # Accept the proposal
              n_ridge[idx[it]] <- j_birth
              n_act[[idx[it]]][j_birth] <- n_act_prop
              feat[[idx[it]]][[j_birth]] <- feat_prop
              knots[[idx[it]]][[j_birth]] <- knots_prop
              proj_dir[[idx[it]]][[j_birth]] <- proj_dir_prop

              if(ridge_type_prop != 'cat'){
                j_quant <- c(j_quant, j_birth)
                n_quant <- n_quant_prop
              }
              ridge_type[j_birth] <- ridge_type_prop

              basis_idx[[j_birth + 1]] <- basis_idx_prop
              n_basis_ridge[j_birth + 1] <- n_basis_prop
              n_basis_total <- n_basis_total + n_basis_prop
              basis_mat <- cbind(basis_mat, ridge_basis_prop)

              # Update weights
              if(adapt_act_feat){
                w_n_act[n_act_prop] <- w_n_act[n_act_prop] + 1
                w_feat[feat_prop] <- w_feat[feat_prop] + 1
              }

              qf_info <- qf_info_prop
              if(it > n_adapt) qf_info <- append_qf_inv_chol(qf_info, dim = n_basis_total)
              sse <- sse_prop
              log_mh_bd <- log_mh_bd_prop
            }
          }
        }
      }else if(move_type == 'death'){ # Death step
        j_death <- sample(n_ridge[idx[it]], 1) # Choose random index to delete

        n_act_prop <- n_act[[idx[it]]][j_death]
        feat_prop <- feat[[idx[it]]][[j_death]]
        if(adapt_act_feat){
          log_mh_act_feat <- log(n_act_max) + log((w_n_act[n_act_prop] - 1)/(sum(w_n_act) - 1))
          w_feat_prop <- w_feat
          w_feat_prop[feat_prop] <- w_feat_prop[feat_prop] - 1
          if(n_act_prop > 1){
            log_mh_act_feat <- log_mh_act_feat + lchoose(p, n_act_prop) + log(dwallenius(w_feat_prop, feat_prop)) # Nott, Kuk, and Duc
          }
        }

        basis_idx_vec_prop <- unlist(basis_idx)[-basis_idx[[j_death + 1]]]
        qf_info_prop <- get_qf_info(BtB[basis_idx_vec_prop, basis_idx_vec_prop, drop = FALSE], Bty[basis_idx_vec_prop])

        n_ridge_prop <- n_ridge[idx[it]] - 1
        if(ridge_type[j_death] == 'cat'){
          n_quant_prop <- n_quant
        }else{
          n_quant_prop <- n_quant - 1
        }

        log_mh_bd_prop <- get_log_mh_bd(n_ridge_prop, n_quant_prop, n_ridge_max)

        if(!is.null(qf_info_prop)){
          if(qf_info_prop$qf < ssy){
            sse_prop <- ssy - c_var_coefs * qf_info_prop$qf
            n_basis_prop <- n_basis_ridge[j_death + 1]

            # Compute acceptance probability
            log_mh <- log_mh_bd - log_mh_bd_prop + log_mh_act_feat + # Adjustment for probability of death proposal
              -n/2 * (log(sse_prop) - log(sse)) + # Part of the marginal likelihood
              log(n_ridge[idx[it]]/n_ridge_mean) # Prior and proposal distribution
            if(prior_coefs == 'zs'){
              log_mh <- log_mh + n_basis_prop/2 * log(var_coefs[idx[it]] + 1) # The rest of the marginal likelihood for Zellner-Siow prior
            }else{
              log_mh <- log_mh - log(10e-6) # The rest of the marginal likelihood for flat prior
            }

            if(log(runif(1)) < log_mh){ # Accept the proposal
              if(ridge_type[j_death] != 'cat'){
                n_quant <- n_quant - 1
                j_quant <- j_quant[j_quant != j_death]
              }
              j_quant[j_quant > j_death] <- j_quant[j_quant > j_death] - 1
              ridge_type <- ridge_type[-j_death]

              basis_mat <- basis_mat[, -basis_idx[[j_death + 1]], drop = FALSE]
              BtB[1:(n_basis_total - n_basis_prop), 1:(n_basis_total - n_basis_prop)] <- BtB[basis_idx_vec_prop, basis_idx_vec_prop, drop = FALSE]
              Bty[1:(n_basis_total - n_basis_prop)] <- Bty[basis_idx_vec_prop]
              if(j_death < n_ridge[idx[it]]){
                for(j in (j_death + 1):n_ridge[idx[it]]){
                  basis_idx[[j + 1]] <- basis_idx[[j + 1]] - n_basis_ridge[j_death + 1]
                }
              }
              basis_idx <- basis_idx[-(j_death + 1)]

              n_basis_ridge <- n_basis_ridge[-(j_death + 1)]
              n_basis_total <- n_basis_total - n_basis_prop
              n_ridge[idx[it]] <- n_ridge[idx[it]] - 1
              n_act[[idx[it]]] <- n_act[[idx[it]]][-j_death]
              feat[[idx[it]]] <- feat[[idx[it]]][-j_death]
              knots[[idx[it]]] <- knots[[idx[it]]][-j_death]
              proj_dir[[idx[it]]] <- proj_dir[[idx[it]]][-j_death]

              # Update weights
              if(adapt_act_feat){
                w_feat <- w_feat_prop
                w_n_act[n_act_prop] <- w_n_act[n_act_prop] - 1
              }

              qf_info <- qf_info_prop
              if(it > n_adapt) qf_info <- append_qf_inv_chol(qf_info, dim = n_basis_total)
              sse <- sse_prop
              log_mh_bd <- log_mh_bd_prop
            }
          }
        }
      }else{ # Change Step
        if(n_quant == 1){
          j_change <- j_quant
        }else{
          j_change <- sample(j_quant, 1) # Which ridge function should we change?
        }

        if(n_act[[idx[it]]][j_change] == 1){
          proj_dir_prop <- matrix(sample(c(-1, 1), 1))
        }else{
          proj_dir_prop <- rps(proj_dir[[idx[it]]][[j_change]], proj_dir_prop_prec) # Get proposed direction
        }

        proj_prop <- X_st[, feat[[idx[it]]][[j_change]], drop = FALSE] %*% proj_dir_prop # Get proposed projection

        if(ridge_type[j_change] == 'cont'){ # Are any variables continuous for this ridge function?
          max_knot0 <- quantile(proj_prop, p_dat_max)
          rg_knot0 <- (max_knot0 -  min(proj_prop)) / prob_relu
          knot0_prop <- max_knot0 - rg_knot0 * runif(1)
          knots_prop <- c(knot0_prop, quantile(proj_prop[proj_prop > knot0_prop], knot_quants)) # Get proposed knots
          ridge_basis_prop <- get_mns_basis(proj_prop, knots_prop) # Get proposed basis function
        }else{
          knots_prop <- NA
          ridge_basis_prop <- proj_prop
        }

        PtP <- t(ridge_basis_prop) %*% ridge_basis_prop # inner product of proposed new basis functions
        BtP <- t(basis_mat) %*% ridge_basis_prop
        Pty <- t(ridge_basis_prop) %*% y

        BtB_prop <- BtB[1:n_basis_total, 1:n_basis_total]
        BtB_prop[basis_idx[[j_change + 1]], ] <- t(BtP)
        BtB_prop[, basis_idx[[j_change + 1]]] <- BtP
        BtB_prop[basis_idx[[j_change + 1]], basis_idx[[j_change + 1]]] <- PtP

        Bty_prop <- Bty[1:n_basis_total]
        Bty_prop[basis_idx[[j_change + 1]]] <- Pty

        qf_info_prop <- get_qf_info(BtB_prop, Bty_prop)

        if(!is.null(qf_info_prop)){
          if(qf_info_prop$qf < ssy){
            sse_prop <- ssy - c_var_coefs * qf_info_prop$qf

            # Compute the acceptance probability
            log_mh <- -n/2 * (log(sse_prop) - log(sse)) # Marginal Likelihood

            if(log(runif(1)) < log_mh){ # Accept the proposal
              knots[[idx[it]]][[j_change]] <- knots_prop
              proj_dir[[idx[it]]][[j_change]] <- proj_dir_prop

              BtB[1:n_basis_total, 1:n_basis_total] <- BtB_prop
              Bty[1:n_basis_total] <- Bty_prop
              basis_mat[, basis_idx[[j_change + 1]]] <- ridge_basis_prop

              qf_info <- qf_info_prop
              if(it > n_adapt) qf_info <- append_qf_inv_chol(qf_info, dim = n_basis_total)
              sse <- sse_prop
            }
          }
        }
      }

      if(it > n_adapt){
        if(it == n_adapt + 1  &&  is.null(qf_info$inv_chol)){
          qf_info <- append_qf_inv_chol(qf_info, dim = n_basis_total)
        }

        sd_resid[idx[it]] <- sqrt(1/rgamma(1, n/2, sse/2))

        # Draw coefs
        coefs[[idx[it]]] <- c_var_coefs * qf_info$ls_est +
          sqrt(c_var_coefs) * sd_resid[idx[it]] * qf_info$inv_chol %*% rnorm(n_basis_total) # Draw coefs
        temp <- qf_info$chol %*% coefs[[idx[it]]]
        qf2 <- c(t(temp) %*% temp)

        if(prior_coefs == 'zs'){
          var_coefs[idx[it]] <- 1/rgamma(1,
                                         shape_var_coefs + n_basis_total/2,
                                         scale_var_coefs + qf2/(2*sd_resid[idx[it]]^2))
          c_var_coefs <- var_coefs[idx[it]] / (var_coefs[idx[it]] + 1)
          sse <- ssy - c_var_coefs * qf_info$qf
        }
      }
    }
  }

  out <- list(n_keep = n_keep, n_ridge = n_ridge,
              n_act = n_act, feat = feat,
              proj_dir = proj_dir, knots = knots,
              coefs = coefs, sd_resid = sd_resid, var_coefs = var_coefs,
              mn_X = mn_X, sd_X = sd_X, X = X, y = y,
              n_ridge_mean = n_ridge_mean, n_ridge_max = n_ridge_max,
              n_act_max = n_act_max, df_spline = df_spline, prob_relu = prob_relu,
              prior_coefs = prior_coefs, shape_var_coefs = shape_var_coefs,
              scale_var_coefs = scale_var_coefs, n_dat_min = n_dat_min,
              scale_proj_dir_prop = scale_proj_dir_prop,
              adapt_act_feat = adapt_act_feat, w_n_act = w_n_act,
              w_feat = w_feat, n_post = n_post, n_burn = n_burn,
              n_adapt = n_adapt, n_thin = n_thin, print_every = print_every,
              bppr_init = bppr_init, call = match.call())

  if(!silent){
    cat(paste0('MCMC iteration ', n_draws, '/', n_draws, ' (', phase, ') ',
               myTimestamp(start_time), ' n ridge: ', n_ridge[idx[n_draws]], '\n'))
  }

  return(structure(out, class = 'bppr'))
}

########################################################################
## prediction function for multivariate response
########################################################################

#' @title BayesPPR Predictions for Multivariate Response
#'
#' @description Predict function for BayesPPR. Outputs the posterior predictive samples for the desired MCMC iterations.
#' @param object a fitted model, output from the \code{bppr} function.
#' @param newdata a matrix of new input values at which to predict.  The columns should correspond to the same variables used in the \code{bppr} function.
#' @param idx_use index of Markov samples to use when generating predictions.
#' @param n_cores number of cores the user desired to utilize for parallel computation.
#' @param par_type a character variable dicatating the type of parallel computation to be performed. Supported values include \code{"fork"}, which uses \code{parallel::mclapply()}, and \code{"socket"}, which uses \code{parallel::parLapply()}.
#' @param ... further arguments passed to or from other methods.
#' @details Bare-bones methods. Could be improved for efficiency.
#' @return An matrix where the first dimension corresponds to MCMC iterations indexed by idx_use, the second dimensions is the same number of rows as \code{newdata}, and the third corresponds to the dimension of the response.
#' @seealso \link{bppr} for model fitting.
#' @export
#' @examples
#' # See examples in bppr documentation.
#'
predict.bppr_pca <-function(object, newdata, idx_use = NULL, n_cores = 1, par_type = 'fork', ...){

  n_keep <- object$fit_list[[1]]$n_keep
  if(is.null(idx_use)){
    idx_use <- 1:n_keep
  }else if(max(idx_use) > n_keep){
    stop("invalid 'idx_use'")
  }
  n_use <- length(idx_use)

  n <- nrow(newdata)
  D <- length(object$pca_Y$mn_Y)

  n_pc <- object$pca_Y$n_pc

  run_predict <- parse(text = 'predict(object$fit_list[[i]], newdata, idx_use = idx_use, ...)')
  run_pca_reverse <- parse(text = 'pca_reverse(preds_Y_new[i, , ], object$pca_Y)')

  if(n_cores == 1){
    preds_Y_new <- array(
      unlist(lapply(1:n_pc, function(i) eval(run_predict))),
      dim = c(n_use, n, n_pc))
    out <- array(
      unlist(lapply(1:n_use, function(i) eval(run_pca_reverse))),
      dim = c(D, n, n_use))
  }else if(par_type == "socket"){
    cl <- parallel::makeCluster(min(n_cores, n_pc, parallel::detectCores()),
                                setup_strategy = "sequential")
    parallel::clusterExport(cl, varlist = c("newdata"), envir = environment())
    preds_Y_new <- array(
      unlist(parallel::parLapply(cl, 1:n_pc, function(i) eval(run_predict))),
      dim = c(n_use, n, n_pc))
    out <- array(
      unlist(parallel::parLapply(cl, 1:n_use, function(i) eval(run_pca_reverse))),
      dim = c(D, n, n_use))
    parallel::stopCluster(cl)
  }else if(par_type == "fork"){
    preds_Y_new <- array(
      unlist(parallel::mclapply(1:n_pc, function(i) eval(run_predict),
                                mc.cores = n_cores)),
      dim = c(n_use, n, n_pc))
    out <- array(
      unlist(parallel::mclapply(1:n_use, function(i) eval(run_pca_reverse),
             mc.cores = n_cores)),
      dim = c(D, n, n_use))
  }

  out <- aperm(out, c(3, 2, 1))
  return(out)
}

get_move_type <- function(n_ridge, n_quant, n_ridge_max){
  if(n_ridge == 0){
    move_type <- 'birth'
  }else if(n_ridge == n_ridge_max){
    if(n_quant == 0){
      move_type <- 'death'
    }else{
      move_type <- sample(c('death', 'change'), 1)
    }
  }else{
    if(n_quant == 0){
      move_type <- sample(c('birth', 'death'), 1)
    }else{
      move_type <- sample(c('birth', 'death', 'change'), 1)
    }
  }

  return(move_type)
}

get_log_mh_bd <- function(n_ridge_prop, n_quant_prop, n_ridge_max){
  if(n_ridge_prop == 0){
    alpha0 <- 0
  }else if(n_ridge_prop == n_ridge_max){
    if(n_quant_prop == 0){
      alpha0 <- 0
    }else{
      alpha0 <- log(2)
    }
  }else{
    if(n_quant_prop == 0){
      alpha0 <- log(2)
    }else{
      alpha0 <- log(3)
    }
  }

  return(alpha0)
}

myTimestamp <- function(start_time = NULL){
  if(is.null(start_time)){
    x <- Sys.time()
    return(paste('--', format(x, "%b %d %X"), '--'))
  }else{
    x <- Sys.time() - start_time
    return(paste('--', format(round(x, 2)), '--'))
  }
}

rps <- function(mu, kappa){ # Get a random draw from the "power spherical" distribution, which has density f(x) \propto (1 + t(mu) %*% x)^kappa
  d <- length(mu)

  uhat <- -mu
  uhat[1] <- uhat[1] + 1
  u <- uhat / sqrt(sum(uhat^2))

  b <- (d - 1)/2
  a <- b + kappa
  z <- rbeta(1, a, b)
  t <- 2*z - 1

  temp <- rnorm(d - 1)
  v <- temp / sqrt(sum(temp^2))
  y <- c(t, sqrt(1 - t^2) * v)

  uy <- sum(u * y)
  x <- y - 2 * u * uy
  return(matrix(x))
}

#' @title BPPR Plot Diagnostics
#'
#' @description Generate diagnostic plots for BPPR model fit.
#' @param x a \code{bppr} object.
#' @param quants quantiles for intervals, if desired.  NULL if not desired.
#' @param pred logical, whether posterior predictions should be plotted (defaults to TRUE).
#' @param ... graphical parameters.
#' @details The first two plots are trace plots for diagnosing convergence.  The third plot is posterior predicted vs observed, with intervals for predictions.  The fourth plot is a histogram of the residuals (of the posterior mean model), with a red curve showing the assumed Normal density (using posterior mean variance). If \code{pred=FALSE} the third and fourth plots are omitted.
#' @return no return value
#' @export
#' @import graphics
#' @seealso \link{bppr}, \link{predict.bppr}
#' @examples
#' # See examples in bppr documentation.
#'
plot.bppr <- function(x, quants = c(.025, .975), pred = TRUE, ...){
  if(!methods::is(x, 'bppr')){
    stop('x must be an object of class bppr')
  }

  op <- par(no.readonly = TRUE)
  on.exit(par(op))
  if(pred){
    par(mfrow = c(2, 2))
  }else{
    par(mfrow = c(1, 2))
  }
  plot(x$n_ridge, type = 'l', ylab = 'Number of Ridge Functions', xlab = 'MCMC Iteration (Post-Burn)')
  plot(x$sd_resid, type = 'l', ylab = 'Residual SD', xlab = 'MCMC Iteration (Post-Burn)')
  if(pred){
    margin <- 2
    preds <- predict(x, x$X)
    yhat <- colMeans(preds)
    if(!is.null(quants)){
      qq <- apply(preds + rnorm(prod(dim(preds)), sd = x$sd_resid), margin, quantile, probs = quants)
      ylim <- range(qq)
      ylab <- 'Interval'
    } else{
      ylim <- range(yhat)
      ylab <- 'Mean'
    }
    plot(x$y, yhat, ylim = ylim, ylab = paste('Posterior Predictive', ylab), xlab = 'Observed Response', main = 'Training Fit', type = 'n', ...)
    if(!is.null(quants)){
      segments(x$y, qq[1,], x$y, qq[2,], col='lightgrey')
    }
    points(x$y, yhat)
    abline(a = 0, b = 1, col = 2)

    hist_dat <- hist(x$y - yhat, plot = FALSE)
    xx <- seq(min(hist_dat$breaks), max(hist_dat$breaks), length.out = 200)
    mn_sd <- mean(x$sd_resid)
    plot(hist_dat, freq = FALSE, col = 'lightgrey', main = 'Posterior Mean Residuals',
         xlab = 'Residuals', ylim = c(0, max(hist_dat$density, dnorm(0, sd = mn_sd))))
    lines(xx, dnorm(xx, sd = mn_sd), col = 2)
  }
  mtext('BPPR Diagnostics', 3, -2, outer=TRUE)
}

library(BayesPPR)
f <- function(xx)
{
  ##########################################################################
  #
  # WELCH ET AL. (1992) FUNCTION
  #
  # Authors: Sonja Surjanovic, Simon Fraser University
  #          Derek Bingham, Simon Fraser University
  # Questions/Comments: Please email Derek Bingham at dbingham@stat.sfu.ca.
  #
  # Copyright 2013. Derek Bingham, Simon Fraser University.
  #
  # THERE IS NO WARRANTY, EXPRESS OR IMPLIED. WE DO NOT ASSUME ANY LIABILITY
  # FOR THE USE OF THIS SOFTWARE.  If software is modified to produce
  # derivative works, such modified software should be clearly marked.
  # Additionally, this program is free software; you can redistribute it
  # and/or modify it under the terms of the GNU General Public License as
  # published by the Free Software Foundation; version 2.0 of the License.
  # Accordingly, this program is distributed in the hope that it will be
  # useful, but WITHOUT ANY WARRANTY; without even the implied warranty
  # of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  # General Public License for more details.
  #
  # For function details and reference information, see:
  # http://www.sfu.ca/~ssurjano/
  #
  ##########################################################################
  #
  # INPUT:
  #
  # xx = c(x1, x2, ..., 20)
  #
  ##########################################################################

  x1  <- xx[1] - 0.5
  x2  <- xx[2] - 0.5
  x3  <- xx[3] - 0.5
  x4  <- xx[4] - 0.5
  x5  <- xx[5] - 0.5
  x6  <- xx[6] - 0.5
  x7  <- xx[7] - 0.5
  x8  <- xx[8] - 0.5
  x9  <- xx[9] - 0.5
  x10 <- xx[10] - 0.5
  x11 <- xx[11] - 0.5
  x12 <- xx[12] - 0.5
  x13 <- xx[13] - 0.5
  x14 <- xx[14] - 0.5
  x15 <- xx[15] - 0.5
  x16 <- xx[16] - 0.5
  x17 <- xx[17] - 0.5
  x18 <- xx[18] - 0.5
  x19 <- xx[19] - 0.5
  x20 <- xx[20] - 0.5

  term1 <- 5*x12 / (1+x1)
  term2 <- 5 * (x4-x20)^2
  term3 <- x5 + 40*x19^3 - 5*x19
  term4 <- 0.05*x2 + 0.08*x3 - 0.03*x6
  term5 <- 0.03*x7 - 0.09*x9 - 0.01*x10
  term6 <- -0.07*x11 + 0.25*x13^2 - 0.04*x14
  term7 <- 0.06*x15 - 0.01*x17 - 0.03*x18

  y <- term1 + term2 + term3 + term4 + term5 + term6 + term7
  return(y)
}


set.seed(71023)
n <- 5000
p <- 20
q <- 2 # 2 categorical inputs
X <- cbind(matrix(runif(n*p), n), sample(0:1, n, replace = TRUE), sample(0:1, n, replace = TRUE))
y <- apply(X, 1, f) + X[, 21] - X[, 22] + 0.5 * X[, 21]*X[, 22] + rnorm(n, sd = 0.2)

fit <- bppr(X, y)
plot(fit$sd_resid, type = 'l')

X_test <- cbind(matrix(runif(n*p), n), sample(0:1, n, replace = TRUE), sample(0:1, n, replace = TRUE))
preds <- predict(fit, X_test)
pred_mn <- apply(preds, 2, mean)
plot(apply(X_test, 1, f), pred_mn)
abline(0, 1, col = 2)
sqrt(mean((apply(X_test, 1, f) - pred_mn)^2))

get_qf_info <- function(BtB, Bty){
  chol_BtB <- tryCatch(chol(BtB), error = function(e) matrix(FALSE))
  if(chol_BtB[1, 1]  &&  !any(is.na(chol_BtB))){
    d <- diag(chol_BtB)
    if(length(d) > 1){
      if((max(d[-1]) / min(d)) > 1000){
        return(NULL)
      }
    }
    bhat <- backsolve(chol_BtB, forwardsolve(chol_BtB, Bty,
                                             transpose = TRUE,
                                             upper.tri = TRUE))
    qf <- t(Bty) %*% bhat
    return(list(chol = chol_BtB, ls_est = bhat, qf = qf))
  }else{
    return(NULL)
  }
}

append_qf_inv_chol <- function(qf_info, dim){
  qf_info$inv_chol <- backsolve(qf_info$chol, diag(dim))
  return(qf_info)
}

#' @title Print BPPR Details
#'
#' @description Print some of the details of a BPPR model.
#' @param x a \code{bppr} object, returned from \code{bppr}.
#' @param ... further arguments passed to or from other methods.
#' @return no return value
#' @export
#'
print.bppr <- function(x, ...){
  cat("\nCall:\n", deparse(x$call), '\n')

  cat("\n              Number of variables: ", ncol(x$X), sep = '')
  cat("\n                      Sample size: ", nrow(x$X), sep = '')

  cat('\n\n')

}


#' @title Summarize BPPR Details
#'
#' @description Summarize some of the details of a BPPR model.
#' @param object a \code{bppr} object, returned from \code{bppr}.
#' @param ... further arguments passed to or from other methods.
#' @return no return value
#' @export
#'
summary.bppr <- function(object,...){
  cat("\nCall:\n", deparse(object$call), '\n')

  cat("\n              Number of variables: ", ncol(object$X), sep = '')
  cat("\n                      Sample size: ", nrow(object$X), sep = '')

  cat("\n\n Number of ridge functions (range):", range(object$n_ridge))
  cat("\n    Posterior mean error sd:", mean(object$sd_resid), '\n\n')

}

test_that("f1 prediction matches previous validated values", {
  cat('Modified Friedman function test (prediction) \n')

  eps <- 1e-12

  fit <- readRDS('../f1_fit.rda') # previous model
  Xtest <- readRDS('../f1_Xtest.rda') # x values
  oldpreds <- readRDS('../f1_preds.rda') # old predictions at x values
  newpreds <- predict(fit, Xtest) # new predictions at x values
  diff <- max(abs(range(newpreds - oldpreds))) # difference between new and old predictions
  expect_that(diff, is_less_than(eps))
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R
\name{summary.bppr}
\alias{summary.bppr}
\title{Summarize BPPR Details}
\usage{
\method{summary}{bppr}(object, ...)
}
\arguments{
\item{object}{a \code{bppr} object, returned from \code{bppr}.}

\item{...}{further arguments passed to or from other methods.}
}
\value{
no return value
}
\description{
Summarize some of the details of a BPPR model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bppr_pca.R
\name{bppr_pca}
\alias{bppr_pca}
\title{BayesPPR for Multivariate Response}
\usage{
bppr_pca(
  X,
  Y,
  n_pc = NULL,
  prop_var = 0.99,
  n_cores = 1,
  par_type = "fork",
  n_ridge_mean = 10,
  n_ridge_max = NULL,
  n_act_max = NULL,
  df_spline = 4,
  prob_relu = 2/3,
  prior_coefs = "zs",
  shape_var_coefs = NULL,
  scale_var_coefs = NULL,
  n_dat_min = NULL,
  scale_proj_dir_prop = NULL,
  adapt_act_feat = TRUE,
  w_n_act = NULL,
  w_feat = NULL,
  n_post = 1000,
  n_burn = 9000,
  n_adapt = 0,
  n_thin = 1,
  print_every = NULL,
  bppr_init_list = NULL
)
}
\arguments{
\item{X}{a data frame or matrix of predictors. Categorical features should be coded as numeric.}

\item{Y}{a data frame or matrix of responses, with columns representing different dimensions of the response.}

\item{n_pc}{number of principle components to be used in pca approximation of Y.}

\item{prop_var}{proportion of variance to be explained by the first \code{n_pc} principle components. Used for automatic selection of \code{n_pc} if \code{is.null(n_pc)}.}

\item{n_cores}{number of cores the user desired to utilize for parallel computation.}

\item{par_type}{a character variable dicatating the type of parallel computation to be performed. Supported values include \code{"fork"}, which uses \code{parallel::mclapply()}, and \code{"socket"}, which uses \code{parallel::parLapply()}.}

\item{n_ridge_mean}{mean for Poisson prior on the number of ridge functions.}

\item{n_ridge_max}{maximum number of ridge functions allowed in the model. Used to avoid memory overload. Defaults to 150 unless the number of observed responses is small.}

\item{n_act_max}{maximum number of active variables in any given ridge function. Defaults to 3 unless categorical features are detected, in which case the default is larger.}

\item{df_spline}{degrees of freedom for spline basis. Stability should be examined for anything other than 4.}

\item{prob_relu}{prior probability that any given ridge function uses a relu transformation.}

\item{prior_coefs}{form of the prior distribution for the basis coefficients. Default is \code{"zs"} for the Zellner-Siow prior. The other option is \code{"flat"}, which is an improper prior.}

\item{shape_var_coefs}{shape for IG prior on the variance of the basis function coefficients. Default is for the Zellner-Siow prior. For the flat, improper prior, \code{shape_var_coefs} is ignored.}

\item{scale_var_coefs}{scale for IG prior on the variance of the basis function coefficients. Default is for the Zellner-Siow prior. For the flat, improper prior, \code{scale_var_coefs} is ignored.}

\item{n_dat_min}{minimum number of observed non-zero datapoints in a ridge function. Defaults to 20 or 0.1 times the number of observations, whichever is smaller.}

\item{scale_proj_dir_prop}{scale parameter for generating proposed projection directions. Should be in (0, 1); default is about 0.002.}

\item{adapt_act_feat}{logical; if \code{TRUE}, use adaptive proposal for feature index sets and number of active features.}

\item{w_n_act}{vector of weights for number of active variables in a ridge function, used in generating proposed basis functions. If \code{adapt_act_feat == FALSE}, it is also used for the prior distribution. Default is \code{rep(1, n_act_max)}.}

\item{w_feat}{vector of weights for feature indices used in a ridge function, used in generating proposed basis functions. If \code{adapt_act_feat == FALSE}, it is also used for the prior distribution. Default is \code{rep(1, ncol(X))}.}

\item{n_post}{number of posterior draws to obtain from the Markov chain after burn-in.}

\item{n_burn}{number of draws to burn before obtaining \code{n_post} draws for inference. If \code{prior_coefs == "flat"} then these are absorbed into the adapt phase.}

\item{n_adapt}{number of adaptive MCMC iterations to perform before burn-in. Skips sampling basis coefficients and residual variance to save time.}

\item{n_thin}{keep every n_thin posterior draws after burn-in.}

\item{print_every}{print the iteration number every print_every iterations. Use \code{print_every = 0} to silence.}

\item{bppr_init_list}{list of length \code{n_pc}, each element containing a list initial values for the Markov chain. Used by \link{bppr_resume}.}
}
\value{
An object of class \code{"bppr_pca"}. Predictions can be obtained by passing the entire object to the \code{predict.bppr_pca} function.
}
\description{
Fits a BayesPPR model to multivariate response using RJMCMC. Can handle categorical features.
}
\details{
Explores BayesPPR model space using RJMCMC. The BayesPPR model has \deqn{y = f(x) + \epsilon,  ~~\epsilon \sim N(0,\sigma^2)} \deqn{f(x) = \beta_0 + \sum_{j=1}^M \beta_j B_j(x)} and \eqn{B_j(x)} is a natural spline basis expansion. We use priors \deqn{\beta \sim N(0,\sigma^2/\tau (B'B)^{-1})} \deqn{M \sim Poisson(\lambda)} as well as the hyper-prior on the variance \eqn{\tau} of the coefficients \eqn{\beta} mentioned in the arguments above.
}
\examples{
# See examples in bppr documentation.

}
\seealso{
\link{predict.bppr_pca} for prediction.
}
\keyword{analysis}
\keyword{component}
\keyword{nonparametric}
\keyword{principle}
\keyword{projection}
\keyword{pursuit}
\keyword{regression}
\keyword{splines}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.bppr}
\alias{plot.bppr}
\title{BPPR Plot Diagnostics}
\usage{
\method{plot}{bppr}(x, quants = c(0.025, 0.975), pred = TRUE, ...)
}
\arguments{
\item{x}{a \code{bppr} object.}

\item{quants}{quantiles for intervals, if desired.  NULL if not desired.}

\item{pred}{logical, whether posterior predictions should be plotted (defaults to TRUE).}

\item{...}{graphical parameters.}
}
\value{
no return value
}
\description{
Generate diagnostic plots for BPPR model fit.
}
\details{
The first two plots are trace plots for diagnosing convergence.  The third plot is posterior predicted vs observed, with intervals for predictions.  The fourth plot is a histogram of the residuals (of the posterior mean model), with a red curve showing the assumed Normal density (using posterior mean variance). If \code{pred=FALSE} the third and fourth plots are omitted.
}
\examples{
# See examples in bppr documentation.

}
\seealso{
\link{bppr}, \link{predict.bppr}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predict.bppr}
\alias{predict.bppr}
\title{BayesPPR Predictions}
\usage{
\method{predict}{bppr}(object, newdata, idx_use = NULL, ...)
}
\arguments{
\item{object}{a fitted model, output from the \code{bppr} function.}

\item{newdata}{a matrix of new input values at which to predict.  The columns should correspond to the same variables used in the \code{bppr} function.}

\item{idx_use}{index of Markov samples to use when generating predictions.}

\item{...}{further arguments passed to or from other methods.}
}
\value{
A matrix with the same number of rows as \code{newdata} and columns corresponding to all MCMC iterations indexed by idx_use. These are samples from the posterior predictive distribution.
}
\description{
Predict function for BayesPPR. Outputs posterior predictive samples for the desired MCMC iterations.
}
\details{
Bare-bones methods. Could be improved for efficiency.
}
\examples{
# See examples in bppr documentation.

}
\seealso{
\link{bppr} for model fitting.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resume.R
\name{bppr_resume}
\alias{bppr_resume}
\title{Resume Sampling of BPPR Parameters}
\usage{
bppr_resume(
  object,
  append = FALSE,
  n_post = 1000,
  n_burn = 9000,
  n_adapt = 0,
  n_thin = 1,
  print_every = 1000
)
}
\arguments{
\item{object}{an object of class \code{"bppr"} containing the values of the current Markov chain, the latest of which constitute initial values for the new chain.}

\item{append}{logical; if \code{TRUE}, new samples will be appended to samples in \code{object}. Otherwise, output will only include new samples.}

\item{n_post}{number of posterior draws to obtain from the Markov chain after burn-in.}

\item{n_burn}{number of draws to burn before obtaining \code{n_post} draws for inference. If \code{prior_coefs == "flat"} then these are absorbed into the adapt phase.}

\item{n_adapt}{number of adaptive MCMC iterations to perform before burn-in. Skips sampling basis coefficients and residual variance to save time.}

\item{n_thin}{keep every n_thin posterior draws after burn-in.}

\item{print_every}{print the iteration number every print_every iterations. Use \code{print_every = 0} to silence.}
}
\value{
An object of class \code{"bppr"}. Predictions can be obtained by passing the entire object to the \code{predict.bppr} function.
}
\description{
Resumes fitting a BayesPPR model where the last Markov chain left off, using the same model parameters.
}
\details{
Continues exploring BayesPPR model space using RJMCMC, beginning at the values contained in the last iteration of \code{object}.
}
\seealso{
\link{predict.bppr} for prediction.
}
\keyword{nonparametric}
\keyword{projection}
\keyword{pursuit}
\keyword{regression}
\keyword{splines}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict_pca.R
\name{predict.bppr_pca}
\alias{predict.bppr_pca}
\title{BayesPPR Predictions for Multivariate Response}
\usage{
\method{predict}{bppr_pca}(object, newdata, idx_use = NULL, n_cores = 1, par_type = "fork", ...)
}
\arguments{
\item{object}{a fitted model, output from the \code{bppr} function.}

\item{newdata}{a matrix of new input values at which to predict.  The columns should correspond to the same variables used in the \code{bppr} function.}

\item{idx_use}{index of Markov samples to use when generating predictions.}

\item{n_cores}{number of cores the user desired to utilize for parallel computation.}

\item{par_type}{a character variable dicatating the type of parallel computation to be performed. Supported values include \code{"fork"}, which uses \code{parallel::mclapply()}, and \code{"socket"}, which uses \code{parallel::parLapply()}.}

\item{...}{further arguments passed to or from other methods.}
}
\value{
An matrix where the first dimension corresponds to MCMC iterations indexed by idx_use, the second dimensions is the same number of rows as \code{newdata}, and the third corresponds to the dimension of the response.
}
\description{
Predict function for BayesPPR. Outputs the posterior predictive samples for the desired MCMC iterations.
}
\details{
Bare-bones methods. Could be improved for efficiency.
}
\examples{
# See examples in bppr documentation.

}
\seealso{
\link{bppr} for model fitting.
}
